# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

from flask_mail import Mail

mail = Mail()


def init_email(self):
    """
    Initialize email facilities.

    :returns: None

    `Flask-Mail <http://pythonhosted.org/Flask-Mail/>`_ is a
    useful tool for creating and sending emails from within a Flask application.
    There are a number of configuration settings beginning with ``MAIL_``
    that permit control over the SMTP credentials used to send email.
    """

    mail.init_app(self.app)



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

import flask


def init_error_handlers(self):
    """
    Initialize handlers for HTTP error events

    :returns: None

    Flask is able to respond to HTTP error codes with custom behaviours.
    By default, it will redirect error 403 (forbidden) to the login page.
    """
    import flask_security as security

    @self.app.errorhandler(403)
    def page_forbidden(e):
        if security.current_user.is_authenticated:
            return flask.redirect(flask.url_for("admin.index"))
        else:
            return flask.redirect(security.url_for_security("login"))


def init_request_handlers(self):
    """
    request handlers

    :returns: None

    Flask handles requests for URLs by scanning the URL path.  Typically,
    any serious functionality will be collected into Views.  However, this
    function is a chance to define a few simple utility URLs.

    If in your application you want to disable the default handlers in
    Flask-Diamond, you can override them like this.

    >>> def request_handlers(self):
    >>>     pass
    """
    @self.app.route('/')
    def index():
        "set up the default handler for requests to /"
        return flask.redirect(flask.url_for("admin.index"))



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

from flask_restful import Api

rest_api = Api()


def init_rest(self, api_map=None):
    """
    Initialize REST API.

    :returns: None

    By default, this function does nothing.  Your application needs to
    overload this function in order to implement your REST API.
    More information about REST can be found in the
    `documentation <http://flask-restful.readthedocs.org/en/latest/>`_.

    api_map is an optional function that can be responsible
    for setting up the API.  This is usually accomplished with a series of
    add_resource() invocations.  api_map must take one parameter, which is
    the Flask-Restful object managed by Flask-Diamond.

    You will end up writing something like this in your application:

    class PlanetResource(Resource):
        def get(self, name):
            planet = Planet.find(name=name)
            if planet:
                return(planet.dump())

    def api_map(rest_extension):
        rest_extension.add_resource(PlanetResource, '/api/planet/<string:name>')

    def create_app():
        application.facet("rest", api_map=api_map)
    """

    if api_map:
        api_map(rest_api)
    rest_api.init_app(self.app)
    return rest_api



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

import flask
import datetime
from flask_security import UserMixin
from flask_security.utils import encrypt_password
from flask_marshmallow.fields import fields
from ..facets.database import db
from ..facets.marshalling import ma
from ..mixins.crud import CRUDMixin
from ..mixins.marshmallow import MarshmallowMixin


"A secondary table is used for the one-to-many relationship: User has many Roles"
roles_users = db.Table('roles_users',
    db.Column('user_id', db.Integer(), db.ForeignKey('user.id')),
    db.Column('role_id', db.Integer(), db.ForeignKey('role.id')))


class UserSchema(ma.Schema):
    confirmed_at = fields.DateTime(required=False)
    last_login_at = fields.DateTime(required=False)
    current_login_at = fields.DateTime(required=False)

    # RoleSchema must be specified in role.py
    roles = fields.Nested('RoleSchema', allow_none=True, many=True)

    class Meta:
        dateformat = ("%F %T %z")
        additional = (
            "id",
            "email",
            "password",
            "active",
            "last_login_ip",
            "current_login_ip",
            "login_count",
        )


class User(db.Model, UserMixin, CRUDMixin, MarshmallowMixin):
    __schema__ = UserSchema

    id = db.Column(db.Integer, primary_key=True)
    "integer -- primary key"

    email = db.Column(db.String(255), unique=True)
    "string -- email address"

    password = db.Column('password', db.String(255), nullable=False)
    "password -- the users's password"

    active = db.Column(db.Boolean())
    "boolean -- whether the user account is active"

    confirmed_at = db.Column(db.DateTime())
    "datetime -- when the user account was confirmed"

    last_login_at = db.Column(db.DateTime())
    "datetime -- the time of the most recent login"

    current_login_at = db.Column(db.DateTime())
    "datetime -- the time of the current login, if any"

    last_login_ip = db.Column(db.String(255))
    "string -- the IP address of the previous login"

    current_login_ip = db.Column(db.String(255))
    "string -- the IP address of the current login"

    login_count = db.Column(db.Integer(), default=0)
    "integer -- the number of times this account been accessed"

    roles = db.relationship('Role',
        enable_typechecks=False,
        secondary=roles_users,
        # backref=db.backref('users', lazy='dynamic'),
    )

    def __str__(self):
        return self.email

    def confirm(self):
        """
        update a User account so that login is permitted

        :returns: None
        """

        self.confirmed_at = datetime.datetime.now()
        self.active = True
        self.save()

    def add_role(self, role_name):
        """
        update a User account so that it includes a new Role

        :param role_name: the name of the Role to add
        :type role_name: string
        """

        from .. import security

        new_role = security.user_datastore.find_or_create_role(role_name)
        security.user_datastore.add_role_to_user(self, new_role)
        db.session.commit()

    @classmethod
    def register(cls, email, password, confirmed=False, roles=None):
        """
        Create a new user account.

        :param email: the email address used to identify the account
        :type email: string
        :param password: the plaintext password for the account
        :type password: string
        :param confirmed: whether to confirm the account immediately
        :type confirmed: boolean
        :param roles: a list containing the names of the Roles for this User
        :type roles: list(string)
        """

        from .. import security

        new_user = security.user_datastore.create_user(
            email=email,
            password=encrypt_password(password)
        )
        db.session.commit()
        if confirmed:
            new_user.confirm()
        if roles:
            for role_name in roles:
                new_user.add_role(role_name)
        flask.current_app.logger.debug("Created user {0}".format(email))
        return new_user



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

from flask_admin import Admin
import flask_security as security
from flask_admin import BaseView, expose, AdminIndexView
from flask_admin.contrib.sqla import ModelView
from flask_admin.base import MenuLink
from flask_security.utils import encrypt_password
import flask

admin = Admin()


def init_administration(self, index_view=None, user=None, role=None):
    """
    Initialize the Administrative GUI.

    :param index_view: the View that will act as the index page of the admin GUI.
    :type index_view: AdminIndexView
    :returns: None

    The administration GUI is substantially derived from `Flask-Admin
    <http://flask-admin.readthedocs.org/en/latest/>`_. When this function
    is called, it will instantiate blueprints so the application serves
    the admin GUI via the URL http://localhost/admin.

    Typically, you will want to call this function even if you override
    it.  The following example illustrates using super() to invoke this
    administration() function from within your own application.

    >>> admin = super(MyApp, self).administration(
    >>>     index_view=MyApp.modelviews.RedirectView(name="Home")
    >>> )
    """

    admin = Admin(
        name=self.app.config["PROJECT_NAME"],
        base_template='admin/login_base.html',
        index_view=index_view or ForceLoginView(name="Home")
    )

    from .. import db

    if not user or not role:
        from ..models.user import User as user
        from ..models.role import Role as role

    admin.init_app(self.app)

    admin.add_view(UserView(user, db.session, category="Admin"))
    admin.add_view(AdminModelView(role, db.session, category="Admin"))

    return admin


class AuthMixin:
    """
    Require user authentication to be accessible
    """

    def is_accessible(self):
        """
        the View is accessible if the User is authenticated
        """

        return security.current_user.is_authenticated


class AdminMixin:
    """
    Require admin Role to be accessible
    """

    def is_accessible(self):
        """
        the View is accessible if the User has the Admin Role
        """

        return security.current_user.has_role("Admin")


class AuthView(AuthMixin, BaseView):
    """
    A View that requires authentication
    """
    pass


class AdminView(AdminMixin, BaseView):
    """
    A View that requires the Admin Role
    """
    pass


class AuthModelView(AuthMixin, ModelView):
    """
    A ModelView that requires authentication
    """
    pass


class AdminModelView(AdminMixin, ModelView):
    """
    A ModelView that requires the Admin Role
    """
    pass


class AuthenticatedMenuLink(AuthMixin, MenuLink):
    """
    A MenuLink that requires authentication
    """
    pass


class UserView(AdminModelView):
    """
    Manage the User Model
    """

    # column_filters = ['email']
    column_exclude_list = ('password', 'active', 'confirmed_at')
    column_searchable_list = ('email', )
    can_delete = False

    create_template = 'admin/create_user.html'

    def create_model(self, form):
        self.model.register(
            name=form.data["name"],
            email=form.data["email"],
            password=form.data["password"],
            confirmed=True,
            roles=["User"],
        )
        return flask.redirect(flask.url_for("user.index_view"))

    def update_model(self, form, model):
        original_password = model.password
        model.update(**form.data)
        if form.data["password"] != original_password:
            model.password = encrypt_password(form.data["password"])
            model.save()
        return flask.redirect(flask.url_for("user.index_view"))


class ForceLoginView(AdminIndexView):
    """
    Allocate the root URL and require authentication
    """

    def is_accessible(self):
        """
        the View is accessible if the User is authenticated
        """

        return security.current_user.is_authenticated

    @expose('/')
    def index(self):
        return self.render("/admin/index.html")



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

from flask_security import SQLAlchemyUserDatastore
from flask_security import Security
from flask import current_app

security = Security()


def init_accounts(self, user=None, role=None, *args, **kwargs):
    """
    Initialize Security for application.

    :param kwargs: parameters that will be passed through to Flask-Security
    :type kwargs: dict
    :returns: None

    A number of common User account operations are provided by `Flask-
    Security <http://pythonhosted.org/Flask-Security/>`_. This function is
    responsible for associating User models in the database with the
    Security object.

    In case you need to override a Flask-Security form (as is the case
    with implementing CAPTCHA) then you must use super() from within your
    application and provide any arguments destined for Flask-Security.

    >>> result = self.super("accounts", user=User, role=Role,
    >>>    confirm_register_form=ExtendedRegisterForm)
    """

    # import database
    from .. import db

    if not user or not role:
        from ..models.user import User as user
        from ..models.role import Role as role

    # create datastore
    user_datastore = SQLAlchemyUserDatastore(db, user, role)
    setattr(Security, "user_datastore", user_datastore)

    security.init_app(self.app, datastore=user_datastore, *args, **kwargs)



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

__project__ = 'Flask-Diamond'
__version__ = '0.5.0'
__author__ = 'Ian Dennis Miller'
__email__ = 'iandennismiller@gmail.com'
__url__ = 'http://flask-diamond.readthedocs.org'
__copyright__ = 'Ian Dennis Miller'



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

from ..facets.database import db


class MarshmallowMixin:

    # dump

    def dump(self):
        "serialize the Model object as a python object"
        return self.__schema__().dump(self).data

    def dumps(self):
        "serialize the Model object as a JSON string"
        return self.__schema__().dumps(self).data

    def dumpf(self, file_handle):
        "write a Model object to file_handle as a JSON string"
        file_handle.write(self.dumps())

    # load

    @classmethod
    def load(cls, python_obj):
        "create a Model object from a python object"
        obj = cls.__schema__().load(python_obj)
        return cls.create(**obj.data)

    @classmethod
    def loads(cls, buf):
        "create a Model object from a JSON-encoded string"
        obj = cls.__schema__().loads(buf)
        return cls.create(**obj.data)

    @classmethod
    def loadf(cls, file_handle):
        "create a Model object from a file_handle pointing to a JSON file"
        return cls.loads(file_handle.read())

    # dump_all

    @classmethod
    def dump_all(cls):
        "write all objects of Model class to an array of python objects"
        return cls.__schema__().dump(cls.query.all(), many=True).data

    @classmethod
    def dumps_all(cls):
        "write all objects of Model class to a JSON-encoded array"
        return cls.__schema__().dumps(cls.query.all(), many=True).data

    @classmethod
    def dumpf_all(cls, file_handle):
        "write all objects of Model class to file_handle as JSON"
        file_handle.write(cls.dumps_all())

    # load_all

    @classmethod
    def load_all(cls, python_objects):
        "create objects of Model class from an array of python objects"
        objs = cls.__schema__().load(python_objects, many=True)
        for obj in objs.data:
            cls.create(_commit=False, **obj)
        db.session.commit()
        db.session.flush()

    @classmethod
    def loads_all(cls, buf):
        "create objects of Model class from a string containing an array of JSON-encoded objects"
        objs = cls.__schema__().loads(buf, many=True)
        for obj in objs.data:
            cls.create(_commit=False, **obj)
        db.session.commit()
        db.session.flush()

    @classmethod
    def loadf_all(cls, file_handle):
        "create objects of Model class from a file containing an array of JSON-encoded objects"
        cls.loads_all(file_handle.read())



"""update to flask-diamond 0.2.0

Revision ID: 20f04b9598da
Revises: cf0f5b45967
Create Date: 2015-02-07 22:54:24.608403

"""

# revision identifiers, used by Alembic.
revision = '20f04b9598da'
down_revision = 'cf0f5b45967'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('user', sa.Column('current_login_at', sa.DateTime(), nullable=True))
    op.add_column('user', sa.Column('current_login_ip', sa.String(length=255), nullable=True))
    op.add_column('user', sa.Column('last_login_at', sa.DateTime(), nullable=True))
    op.add_column('user', sa.Column('last_login_ip', sa.String(length=255), nullable=True))
    op.add_column('user', sa.Column('login_count', sa.Integer(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('user', 'login_count')
    op.drop_column('user', 'last_login_ip')
    op.drop_column('user', 'last_login_at')
    op.drop_column('user', 'current_login_ip')
    op.drop_column('user', 'current_login_at')
    ### end Alembic commands ###



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

from .. import Diamond
from nose.plugins.attrib import attr
from unittest import TestCase


class MyDiamondApp(Diamond):
    def init_blueprints(self):
        self.super("blueprints")


def create_app():
    diamond = Diamond()
    diamond.facet("configuration")
    diamond.facet("logs")
    diamond.facet("database")
    diamond.facet("marshalling")
    diamond.facet("accounts")
    diamond.facet("blueprints")
    diamond.facet("signals")
    diamond.facet("forms")
    diamond.facet("error_handlers")
    diamond.facet("request_handlers")
    diamond.facet("administration")
    diamond.facet("rest")
    diamond.facet("webassets")
    diamond.facet("email")
    diamond.facet("debugger")
    diamond.facet("task_queue")
    return diamond.app


class BasicTestCase(TestCase):

    @attr("single")
    def test_create_app(self):
        from ..facets.database import db
        from ..models.role import Role
        from ..models.user import User

        app = create_app()
        with app.app_context():
            db.drop_all()
            try:
                db.create_all()

                Role.add_default_roles()

                User.register(
                    email="guest@example.com",
                    password="guest",
                    roles=["User"],
                )

                User.register(
                    email="admin@example.com",
                    password="abc",
                    roles=["Admin"],
                )

                self.assertIsNotNone(app, msg="create app")
            finally:
                db.drop_all()



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

from flask_security.signals import user_registered
from .accounts import security
from .database import db


def init_signals(self):
    """
    Initialize Flask signal handlers

    :returns: None

    Flask provides a number of signals corresponding to things that happen
    during the operation of the application, which can also be thought of
    as events.  It is possible to create signal handlers that will respond
    to these events with some behaviour.
    """
    @user_registered.connect_via(self.app)
    def user_registered_sighandler(sender, **extra):
        "add User role to all self-registration users"
        user_role = security.user_datastore.find_role("User")
        security.user_datastore.add_role_to_user(extra['user'], user_role)
        db.session.commit()
        self.app.logger.info("added role User to %s" % extra['user'])



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

import re
import os
import codecs
from setuptools import setup, find_packages


def read(*rnames):
    return codecs.open(os.path.join(os.path.dirname(__file__), *rnames), 'r', 'utf-8').read()


def grep(attrname):
    pattern = r"{0}\W*=\W*'([^']+)'".format(attrname)
    strval, = re.findall(pattern, read('flask_diamond/__meta__.py'))
    return strval


setup(
    version=grep('__version__'),
    name='Flask-Diamond',
    description=("Flask-Diamond provides some opinions about "
        "data-centric Internet applications and systems. "
        "Flask-Diamond is a batteries-included Flask framework, "
        "sortof like Django but radically decomposable."
    ),
    packages=find_packages(),
    scripts=[
        "bin/flask-diamond",
    ],
    long_description=read('Readme.rst'),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Internet :: WWW/HTTP",
    ],
    include_package_data=True,
    keywords='',
    author=grep('__author__'),
    author_email=grep('__email__'),
    url=grep('__url__'),
    install_requires=read('requirements.txt'),
    license='MIT',
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# Flask-Diamond (c) Ian Dennis Miller

import flask
from .facets import *

try:
    from flask import _app_ctx_stack as stack
except ImportError:
    from flask import _request_ctx_stack as stack

application = None


class Diamond:
    """
    A Diamond application.

    :param app: a Flask app that you created on your own
    :type app: Flask
    :returns: None
    """

    def __init__(self, name=None):
        """
        Initialize a Diamond application.

        :param app: a Flask app that you created on your own
        :type app: Flask
        :returns: Flask -- the initialized application object

        This function is the backbone of every Diamond application.  It will
        initialize every component of the application.  To control the
        behaviour of the initialization process, override these functions
        within your own application.
        """

        if not name:
            name = __name__

        self.app = flask.Flask(name)

        if hasattr(self.app, 'teardown_appcontext'):
            self.app.teardown_appcontext(self.teardown)
        else:
            self.app.teardown_request(self.teardown)

    def facet(self, extension_name, *args, **kwargs):
        """
        initialize an extension
        """
        init_method = "init_{0}".format(extension_name)
        if not hasattr(self, init_method):
            method_to_call = globals()[init_method]
        else:
            method_to_call = getattr(self, init_method)
        setattr(self, init_method, method_to_call)

        try:
            # try to explicitly pass self as the first parameter
            result = method_to_call(self, *args, **kwargs)
        except TypeError:
            # just call it because it will be wrapped to inject self
            result = method_to_call(*args, **kwargs)

        self.app.logger.debug("facet {0}".format(extension_name))
        return result

    def super(self, extension_name, *args, **kwargs):
        """
        invoke the initialization method for the superclass

        ex: self.super("administration")
        """

        init_method = "init_{0}".format(extension_name)
        # ensure the global version is called
        method_to_call = globals()[init_method]
        result = method_to_call(self, *args, **kwargs)
        return result

    def teardown(self, exception):
        """
        Remove any persistent connections during application context teardown.

        :returns: None
        """

        ctx = stack.top
        if hasattr(ctx, 'diamond'):
            pass
            # ctx.sqlite3_db.close()

    @property
    def _app(self):
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'app'):
                pass
            return ctx.app
